import { EventProperties, Traits, CoreAnalytics, CoreContext, CorePlugin, CoreSegmentEvent, Integrations } from '@segment/analytics-core';
import { AnalyticsSettings } from './settings';
import { Callback } from './dispatch-emit';
import { NodeEmitter } from './emitter';
export declare class Context extends CoreContext {
}
export interface Plugin extends CorePlugin {
}
declare type Timestamp = string | Date;
/**
 * An ID associated with the user. Note: at least one of userId or anonymousId must be included.
 **/
declare type IdentityOptions = {
    userId: string;
    anonymousId?: string;
} | {
    userId?: string;
    anonymousId: string;
};
/**
 * A dictionary of extra context to attach to the call.
 * Note: context differs from traits because it is not attributes of the user itself.
 */
declare type AdditionalContext = Record<string, any>;
declare type SegmentEventType = 'track' | 'page' | 'identify' | 'alias' | 'screen';
export interface SegmentEvent extends CoreSegmentEvent {
    type: SegmentEventType;
}
export declare class Analytics extends NodeEmitter implements CoreAnalytics {
    private readonly _eventFactory;
    private _isClosed;
    private _pendingEvents;
    private readonly _closeAndFlushDefaultTimeout;
    private readonly _publisher;
    private readonly _queue;
    ready: Promise<void>;
    constructor(settings: AnalyticsSettings);
    get VERSION(): string;
    /**
     * Call this method to stop collecting new events and flush all existing events.
     * This method also waits for any event method-specific callbacks to be triggered,
     * and any of their subsequent promises to be resolved/rejected.
     */
    closeAndFlush({ timeout, }?: {
        /** Set a maximum time permitted to wait before resolving. */
        timeout?: number;
    }): Promise<void>;
    private _dispatch;
    /**
     * Combines two unassociated user identities.
     * @link https://segment.com/docs/connections/sources/catalog/libraries/server/node/#alias
     */
    alias({ userId, previousId, context, timestamp, integrations, }: {
        userId: string;
        previousId: string;
        context?: AdditionalContext;
        timestamp?: Timestamp;
        integrations?: Integrations;
    }, callback?: Callback): void;
    /**
     * Associates an identified user with a collective.
     *  @link https://segment.com/docs/connections/sources/catalog/libraries/server/node/#group
     */
    group({ timestamp, groupId, userId, anonymousId, traits, context, integrations, }: IdentityOptions & {
        groupId: string;
        traits?: Traits;
        context?: AdditionalContext;
        timestamp?: Timestamp;
        integrations?: Integrations;
    }, callback?: Callback): void;
    /**
     * Includes a unique userId and (maybe anonymousId) and any optional traits you know about them.
     * @link https://segment.com/docs/connections/sources/catalog/libraries/server/node/#identify
     */
    identify({ userId, anonymousId, traits, context, integrations, }: IdentityOptions & {
        traits?: Traits;
        context?: AdditionalContext;
        integrations?: Integrations;
    }, callback?: Callback): void;
    /**
     * The page method lets you record page views on your website, along with optional extra information about the page being viewed.
     * @link https://segment.com/docs/connections/sources/catalog/libraries/server/node/#page
     */
    page({ userId, anonymousId, category, name, properties, context, timestamp, integrations, }: IdentityOptions & {
        category?: string;
        name?: string;
        properties?: EventProperties;
        timestamp?: Timestamp;
        context?: AdditionalContext;
        integrations?: Integrations;
    }, callback?: Callback): void;
    /**
     * Records screen views on your app, along with optional extra information
     * about the screen viewed by the user.
     *
     * TODO: This is not documented on the segment docs ATM (for node).
     */
    screen({ userId, anonymousId, category, name, properties, context, timestamp, integrations, }: Parameters<Analytics['page']>[0], callback?: Callback): void;
    /**
     * Records actions your users perform.
     * @link https://segment.com/docs/connections/sources/catalog/libraries/server/node/#track
     */
    track({ userId, anonymousId, event, properties, context, timestamp, integrations, }: IdentityOptions & {
        event: string;
        properties?: EventProperties;
        context?: AdditionalContext;
        timestamp?: Timestamp;
        integrations?: Integrations;
    }, callback?: Callback): void;
    /**
     * Registers one or more plugins to augment Analytics functionality.
     * @param plugins
     */
    register(...plugins: CorePlugin<any, any>[]): Promise<void>;
    /**
     * Deregisters one or more plugins based on their names.
     * @param pluginNames - The names of one or more plugins to deregister.
     */
    deregister(...pluginNames: string[]): Promise<void>;
}
export {};
//# sourceMappingURL=analytics-node.d.ts.map